# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0

"""
The energyml.utils module.
It contains tools for energyml management.

Please check the following module (depending on your need):
    - energyml-opc
    - energyml-common2-0
    - energyml-common2-1
    - energyml-common2-2
    - energyml-common2-3
    - energyml-resqml2-0-1
    - energyml-resqml2-2-dev3
    - energyml-resqml2-2
    - energyml-witsml2-0
    - energyml-witsml2-1
    - energyml-prodml2-0
    - energyml-prodml2-2
"""
