# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
"""
The data module.

Contains functions to help the read of specific entities like Grid2DRepresentation, TriangulatedSetRepresentation etc.
It also contains functions to export data into OFF/OBJ format.
"""
